import 'dart:math' as math;
import 'vector.dart';
import 'rectangle.dart';

final RNG rng = RNG(DateTime.now().millisecondsSinceEpoch);

class RNG {
  math.Random _randomGenerator;

  RNG([int seed]) : _randomGenerator = math.Random(seed);

  set seed(int value) => _randomGenerator = math.Random(value);

  ///Return a random generated integer. If [maxBound] isn't
  ///provided the returned value will be in range [0, [bound]).
  ///Otherwise, if [maxBound] is provided, in range
  ///[[bound], [maxBound]).
  int range(int bound, [int maxBound]) {
    if (maxBound == null) return _randomGenerator.nextInt(bound);

    return bound + _randomGenerator.nextInt(maxBound);
  }

  ///Return a random generated integer. If [maxBound] isn't
  ///provided the returned value will be in range [0, [bound]].
  ///Otherwise, if [maxBound] is provided, in range
  ///[[bound], [maxBound]].
  int inclusiveRange(int bound, [maxBound]) {
    if (maxBound == null) return range(bound + 1);

    return range(bound, maxBound + 1);
  }

  ///Return a random generated floting point number. If no
  ///paramethers are provided the returned number is in range
  ///`[0, 1]`; if only [bound] is provided in range `[0, bound]`;
  ///otherwise, if [bound] and [maxBound] are provided
  ///in range `[[bound], [maxBound]]`.
  double float([double bound, double maxBound]) {
    if (bound == null)
      return _randomGenerator.nextDouble();
    else if (maxBound == null)
      return _randomGenerator.nextDouble() * bound;
    else
      return _randomGenerator.nextDouble() * (maxBound - bound) + bound;
  }

  ///Return `true` with [chance] percent.
  bool percent(int chance) => range(100) < chance;

  ///Return `true` one in [value] times.
  ///*WATCHOUT* always returns `true` with
  ///[value] = `0`.
  bool oneIn(int value) => range(value) == 0;

  ///Return a random item from the given List [items].
  ///The given list won't be modified.
  T select<T>(List<T> items) => items[range(items.length)];

  ///Pop a random item from the given List [items].
  ///Items' order won't change.
  T popOrdered<T>(List<T> items) => items.removeAt(range(items.length));

  ///Pop a random item from the given List [items].
  ///The popped item is replaced by the last item
  ///in the list.
  T pop<T>(List<T> items) {
    int i = range(items.length);
    T poppedItem = items.removeAt(i);

    items[i] = items.last;
    items.removeLast();

    return poppedItem;
  }

  ///Shaffle given list [items]
  void shaffle<T>(List<T> items) => items.shuffle(_randomGenerator);

  ///Return random floating point number with
  ///a normal distribution
  ///ref: https://en.wikipedia.org/wiki/Marsaglia_polar_method
  double normal() {
    double u, v, s;

    do {
      u = float(-1.0, 1.0);
      v = float(-1.0, 1.0);
      s = u * u + v * v;
    } while (s >= 1.0 || s == 0.0);

    return math.sqrt(-2.0 * math.log(s) / s) * u;
  }

  ///Return a random integer in the range
  ///[[center] - [range]], [[center] + [range]] but with a
  ///greater distribution toward [center].
  int triangle(int center, int range) {
    int x = inclusiveRange(range);
    int y = inclusiveRange(range);

    if (x <= y) return center + x;

    return center + x - range - 1;
  }

  ///Returns a random vector within given [rect]
  Vector boundedVector(Rectangle rect) =>
      Vector(range(rect.left, rect.right), range(rect.top, rect.bottom));
}
